package org.hardknockjava.finale.driver;

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;

import javax.imageio.ImageIO;

import org.apache.pdfbox.pdmodel.PDDocument;
import org.apache.pdfbox.pdmodel.PDPage;
import org.apache.pdfbox.pdmodel.PDPageContentStream;
import org.apache.pdfbox.pdmodel.common.PDRectangle;
import org.apache.pdfbox.pdmodel.font.PDFont;
import org.apache.pdfbox.pdmodel.font.PDType1Font;
import org.apache.pdfbox.pdmodel.font.Standard14Fonts;
import org.apache.pdfbox.pdmodel.graphics.image.PDImageXObject;
import org.hardknockjava.finale.grid.Clue;
import org.hardknockjava.finale.grid.Direction;
import org.hardknockjava.finale.grid.Grid;
import org.hardknockjava.finale.grid.Square;

public class CrosswordPdfWriter {
    private Crossword crossword;
    private List<String> clueList;
    private float margin = 72;

    // Maintain a list of rectangles occupied by the photograph
    // and the grid.
    private List<PDRectangle> restrictedList = new ArrayList<PDRectangle>();

    // squareSide is the number of points in one side of a grid square.
    private static final float SQUARE_SIDE = 20;
    private static final float IMAGE_WIDTH_IN_POINTS = 108;

    public static void main(String[] args) throws Exception {
        Crossword crossword = new Crossword();
        CrosswordPdfWriter printer = new CrosswordPdfWriter(crossword);
        printer.printPuzzle();
        System.out.println("Done");
    }

    public CrosswordPdfWriter(Crossword crossword) {
        this.crossword = crossword;

        // Copy the clues from crossword to clueList, but insert
        // "ACROSS" for the first entry and add "DOWN" before
        // the first down word appears.
        boolean downFound = false;
        crossword.getClueList().sort(null);
        clueList = new ArrayList<String>();
        clueList.add("ACROSS");
        for (Clue clue : crossword.getClueList()) {
            if (clue.getDirection().equals(Direction.DOWN)) {
                if (!downFound) {
                    clueList.add("");
                    clueList.add("DOWN");
                    downFound = true;
                }
            }
            clueList.add(clue.getNumber() + ". " + clue.getText());
        }
    }

    public void printPuzzle() throws Exception {
        PDDocument doc = new PDDocument();
        printPage1(doc);
        printPage2(doc);

        Date today = new Date();
        SimpleDateFormat format = new SimpleDateFormat("yyyyMMdd-HH24mmss");
        String fileName = "crossword-" + format.format(today) + ".pdf";
        File crosswordDir = new File("pdf");
        crosswordDir.mkdirs();
        File crosswordFile = new File(crosswordDir, fileName);
        crosswordFile.delete();
        doc.save(crosswordFile);
        doc.close();
    }

    private void printPage1(PDDocument doc) throws Exception {
        PDPage page = new PDPage();
        doc.addPage(page);
        PDPageContentStream contentStream = new PDPageContentStream(doc, page);
        try {
            printPage1(doc, page, contentStream);
        } finally {
            contentStream.close();
        }
    }

    private void printPage1(PDDocument doc, PDPage page, PDPageContentStream contentStream) throws Exception {
        // Add the photograph to the page. Scale it so
        // its width is IMAGE_WIDTH.
        float heightToWidth = ((float) crossword.getImage().getHeight()) / ((float) (crossword.getImage().getWidth()));
        float targetImageHeight = IMAGE_WIDTH_IN_POINTS * heightToWidth;
        ByteArrayOutputStream os = new ByteArrayOutputStream();
        ImageIO.write(crossword.getImage(), "jpg", os);
        byte[] imageArray = os.toByteArray();
        PDImageXObject pdImage = PDImageXObject.createFromByteArray(doc, imageArray, "spirit");

        // The starting y coordinate is the height of the print
        // area less originY.
        PDRectangle box = page.getMediaBox();
        float imageOriginY = (box.getHeight() - margin - pdImage.getHeight()) + SQUARE_SIDE;
        contentStream.drawImage(pdImage, margin, imageOriginY, IMAGE_WIDTH_IN_POINTS, targetImageHeight);

        // Compute where the crossword grid will begin horizontally.
        float imageToGridGap = 6;
        float gridOriginX = margin + IMAGE_WIDTH_IN_POINTS + imageToGridGap;
        float pageOriginY = box.getHeight() - margin;
        printGrid(page, contentStream, gridOriginX, pageOriginY, false);

        // Add the rectangles occupied by the grid and the photograph
        // to restrictedList.
        restrictedList.add(new PDRectangle(gridOriginX, pageOriginY, //
                SQUARE_SIDE * crossword.getGrid().getColCount(), SQUARE_SIDE * crossword.getGrid().getRowCount()));
        restrictedList.add(new PDRectangle(0, pageOriginY, //
                IMAGE_WIDTH_IN_POINTS, targetImageHeight));

        // Repeat the process of trying to fit all the clues on
        // the page, trying successively smaller fonts until
        // successful.
        float fontSize = 10;
        while (true) {
            if (addClues(fontSize, contentStream, restrictedList, page)) {
                break;
            }
            fontSize--;
        }
    }

    private void printPage2(PDDocument doc) throws Exception {
        PDPage page = new PDPage();
        doc.addPage(page);
        PDPageContentStream contentStream = new PDPageContentStream(doc, page);

        PDRectangle box = page.getMediaBox();
        float gridOriginY = box.getHeight() - margin;

        try {
            printGrid(page, contentStream, margin, gridOriginY, true);
        } finally {
            contentStream.close();
        }
    }

    private void printGrid(PDPage page, PDPageContentStream contentStream, float gridOriginX, float gridOriginY,
            boolean printSolution) throws Exception {
        printGridSquares(page, contentStream, gridOriginX, gridOriginY);
        printGridContent(page, contentStream, gridOriginX, gridOriginY, printSolution);
    }

    private void printGridSquares(PDPage page, PDPageContentStream contentStream, float gridOriginX, float gridOriginY)
            throws Exception {
        Grid grid = crossword.getGrid();

        float y = gridOriginY;

        for (int row = 0; row < grid.getRowCount(); row++) {
            float x = gridOriginX;
            for (int column = 0; column < grid.getColCount(); column++) {
                Square square = grid.getSquare()[row][column];
                contentStream.addRect(x, y, SQUARE_SIDE, SQUARE_SIDE);
                if (square.getValue() == null) {
                    contentStream.fillAndStroke();
                } else {
                    contentStream.stroke();
                }

                // Move x further to the right.
                x = x + SQUARE_SIDE;
            }

            // Move y further down.
            y = y - SQUARE_SIDE;
        }
    }

    private void printGridContent(PDPage page, PDPageContentStream contentStream, float gridOriginX, float gridOriginY,
            boolean printSolution) throws Exception {
        float fontSize;
        PDFont font;
        if (printSolution) {
            font = new PDType1Font(Standard14Fonts.FontName.HELVETICA_BOLD);
            fontSize = 10;
        } else {
            font = new PDType1Font(Standard14Fonts.FontName.HELVETICA);
            fontSize = 5;
        }

        Grid grid = crossword.getGrid();

        // Set the y coordinate to the height of the page less originY
        // less the font size.
        float y = gridOriginY;

        for (int row = 0; row < grid.getRowCount(); row++) {
            float x = gridOriginX;
            for (int column = 0; column < grid.getColCount(); column++) {
                Square square = grid.getSquare()[row][column];
                if (square.getValue() != null) {
                    contentStream.beginText();
                    contentStream.setFont(font, fontSize);
                    if (printSolution) {
                        String value = square.getValue().toString();
                        float charWidth = font.getStringWidth(value) / 1000.0F;
                        float newX = x + ((SQUARE_SIDE - charWidth) / 2) - 2;
                        contentStream.newLineAtOffset(newX, y + 5);
                        contentStream.showText(value);
                    } else {
                        if (square.getClueNumber() != null) {
                            contentStream.newLineAtOffset(x + 1, y + SQUARE_SIDE - fontSize - 1);
                            contentStream.showText(square.getClueNumber().toString());
                        }
                    }
                    contentStream.endText();
                }

                x = x + SQUARE_SIDE;
            }

            // Move vertically one line.
            y = y - SQUARE_SIDE;
        }
    }

    // Add the clues to the grid, if they fit. If they do, return true.
    private boolean addClues(float fontSize, PDPageContentStream contentStream, List<PDRectangle> restrictedList,
            PDPage page) throws Exception {
        PDFont clueFont = new PDType1Font(Standard14Fonts.FontName.TIMES_ROMAN);
        contentStream.setFont(clueFont, fontSize);

        // Start by dividing the page into five columns.
        int columnCount = 5;
        float columnWidth = 72F * 7.5F / columnCount;

        // For each column, determine how much space there is between
        // any restricted area, or the top of the page, and the bottom
        // of the page. Then add the number of lines that will fit to
        // a running total. We won't try to put clues between the two
        // graphics.
        int linesAvailable = 0;
        float lineHeight = ((clueFont.getBoundingBox().getHeight()) / 1000F * fontSize);

        for (int column = 0; column < columnCount; column++) {
            float x = column * columnWidth;
            float columnTop = clueColumnStartY(x, columnWidth, lineHeight, restrictedList, page);
            float spaceAvailable = columnTop - margin;
            Float linesInColumn = spaceAvailable / lineHeight;
            linesAvailable += linesInColumn.intValue();
        }

        // Now see if the clues will fit.
        List<String> clueTextList = new ArrayList<String>();
        for (String clueText : clueList) {
            // Split the clue text into as many lines as necessary
            // and add them to clueTextList. We'll leave a cushion
            // of 7 points between columns.
            // If divideClue couldn't succeed, return false.
            List<String> dividedClues = divideClue(clueText, clueFont, fontSize, columnWidth - 7F);
            if (dividedClues.isEmpty()) {
                return false;
            }
            clueTextList.addAll(dividedClues);
        }

        // If the number if lines required exceeds the number available,
        // return false. The caller will try again with a smaller font.
        if (clueTextList.size() > linesAvailable) {
            return false;
        }

        // The lines will fit in the available space. Add them
        // to the page.

        // Determine the starting y for the first column
        // from x and the values in restrictedList.
        // Increment y after each line. When the next
        // y would exceed the page bottom, add columnWidth
        // to x and compute y for the top of the next column.
        float x = margin - columnWidth;
        float y = -1F;
        boolean textMode = false;
        contentStream.setLeading(lineHeight);
        while (!clueTextList.isEmpty()) {
            if (y < margin) {
                if (textMode) {
                    contentStream.endText();
                }
                contentStream.beginText();
                textMode = true;
                x += columnWidth;
                y = clueColumnStartY(x, columnWidth, lineHeight, restrictedList, page);
                contentStream.newLineAtOffset(x, y);
            } else {
                contentStream.newLine();
            }
            String clueText = clueTextList.remove(0);
            contentStream.showText(clueText);
            y -= lineHeight;
        }
        if (textMode) {
            contentStream.endText();
        }

        return true;
    }

    // Get the y position for the first clue in a column.
    private float clueColumnStartY(float x, float columnWidth, float lineHeight, List<PDRectangle> restrictedList,
            PDPage page) {
        PDRectangle box = page.getMediaBox();
        float columnTop = box.getHeight() - margin;
        for (PDRectangle rectangle : restrictedList) {
            // If x or (x + columnWidth) is between the minimum
            // and maximum x for the rectangle, ensure that
            // columnTop is below the bottom of the rectangle.
            boolean xContained = false;
            if (x >= rectangle.getLowerLeftX() && x <= rectangle.getUpperRightX()) {
                xContained = true;
            }
            if ((x + columnWidth) >= rectangle.getLowerLeftX() && (x + columnWidth) <= rectangle.getUpperRightX()) {
                xContained = true;
            }
            if (xContained) {
                float lowerBound = rectangle.getLowerLeftY() - rectangle.getHeight();
                if (lowerBound <= columnTop) {
                    // Calculate columnTop as the bottom of the rectangle
                    // plus 0.25 in.
                    columnTop = lowerBound - 18F;
                }
            }
        }

        return columnTop;
    }

    // Split the clue text into as many lines as necessary to fit within
    // columnWidth.

    private List<String> divideClue(String clueText, PDFont clueFont, float fontSize, float columnWidth)
            throws Exception {
        List<String> lines = new ArrayList<String>();
        StringBuilder textBuf = new StringBuilder(clueText);

        // While the width of the clue exceeds the available width,
        // break off words until the next word would exceed the width.
        while (textWidth(textBuf.toString(), clueFont, fontSize) > columnWidth) {
            // If there's a space left over at the start of the buffer,
            // remove it.
            if (textBuf.charAt(0) == ' ') {
                textBuf.deleteCharAt(0);
            }

            // lineText contains the words as they're added
            // to the line.
            StringBuilder lineText = new StringBuilder();

            // Start each line after the first with a space.
            if (!lines.isEmpty()) {
                lineText.append(" ");
            }

            boolean full = false;
            String space = "";

            while (!full) {
                // Find the next space following the first character.
                int spacePos = textBuf.indexOf(" ", 1);

                // If there was no space, set spacePos to the
                // length of textBuf.
                if (spacePos < 0) {
                    spacePos = textBuf.length();
                }

                // The next word is everything up to but not
                // including the space.
                String nextWord = textBuf.substring(0, spacePos);

                // If the current lineText plus a space plus
                // nextWord exceeds the column width, we're
                // done.
                // Otherwise, append nextWord to lineText and
                // delete it from textBuf.
                if (textWidth(lineText.toString() + " " + nextWord, clueFont, fontSize) > columnWidth) {
                    full = true;
                } else {
                    lineText.append(space + nextWord);
                    textBuf.delete(0, spacePos + 1);
                    space = " ";
                }
            }

            // If the text to be added is empty, return an empty list.
            if (lineText.toString().isBlank()) {
                lines.clear();
                return lines;
            }
            // Add the accumulated lineText to the return list.
            lines.add(lineText.toString());
        }

        // When what's left of the clue fits within columnWidth,
        // add it to lines.
        if (!lines.isEmpty()) {
            textBuf.insert(0, ' ');
        }
        lines.add(textBuf.toString());

        return lines;
    }

    private float textWidth(String text, PDFont clueFont, float fontSize) throws Exception {
        return clueFont.getStringWidth(text.toString()) / 1000 * fontSize;
    }
}
