package org.hardknockjava.finale.driver;

import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Graphics;
import java.awt.Rectangle;
import java.awt.print.PageFormat;
import java.awt.print.Paper;
import java.awt.print.Printable;
import java.awt.print.PrinterException;
import java.awt.print.PrinterJob;
import java.util.ArrayList;
import java.util.List;

import org.hardknockjava.finale.grid.Clue;
import org.hardknockjava.finale.grid.Direction;
import org.hardknockjava.finale.grid.Grid;
import org.hardknockjava.finale.grid.Square;

public class CrosswordPrinter implements Printable {
    private static final int DEFAULT_MARGIN = 72;
    private static final int IMAGE_WIDTH_IN_POINTS = 108;
    private Crossword crossword;
    private List<String> clueList;

    // squareSide is the number of points in one side of a grid square.
    private static final int SQUARE_SIDE = 20;

    public static void main(String[] args) throws Exception {
        Crossword crossword = new Crossword();
        CrosswordPrinter printer = new CrosswordPrinter(crossword);
        printer.printPuzzle();

        crossword.writeToPdf();

        System.out.println("Done");
    }

    public CrosswordPrinter(Crossword crossword) {
        this.crossword = crossword;

        // Copy the clues from crossword to clueList, but insert
        // "ACROSS" for the first entry and add "DOWN" before
        // the first down word appears.
        crossword.getClueList().sort(null);
        boolean downFound = false;
        crossword.getClueList().sort(null);
        clueList = new ArrayList<String>();
        clueList.add("ACROSS");
        for (Clue clue : crossword.getClueList()) {
            if (clue.getDirection().equals(Direction.DOWN)) {
                if (!downFound) {
                    clueList.add("");
                    clueList.add("DOWN");
                    downFound = true;
                }
            }
            clueList.add(clue.getNumber() + ". " + clue.getText());
        }
    }

    public void printPuzzle() throws Exception {
        PrinterJob job = PrinterJob.getPrinterJob();
        job.setPrintable(this);

        if (!job.printDialog()) {
            return;
        }

        job.print();
    }

    @Override
    public int print(Graphics graphics, PageFormat pageFormat, int pageIndex) throws PrinterException {
        // Ensure there's at least a one-inch margin on each edge.
        double margin = pageFormat.getImageableX();
        {
            Paper paper = pageFormat.getPaper();

            if (pageFormat.getImageableX() < DEFAULT_MARGIN || pageFormat.getImageableY() < DEFAULT_MARGIN) {
                margin = DEFAULT_MARGIN;
            }

            paper.setImageableArea(margin, margin, paper.getWidth() - margin * 2, //
                    paper.getHeight() - margin * 2);
            pageFormat.setPaper(paper);
        }

        int imageableX = ((Double) pageFormat.getImageableX()).intValue();
        int imageableY = ((Double) pageFormat.getImageableY()).intValue();
        graphics.translate(imageableX, imageableY);

        // Maintain a list of rectangles occupied by the photograph
        // and the grid.
        List<Rectangle> restrictedList = new ArrayList<Rectangle>();

        switch (pageIndex) {
        case 0: {
            // Add the photograph to the page. Scale it so
            // its width is IMAGE_WIDTH.
            float heightToWidth = ((float) crossword.getImage().getHeight())
                    / ((float) (crossword.getImage().getWidth()));
            int targetImageHeight = (int) (IMAGE_WIDTH_IN_POINTS * heightToWidth);

            graphics.drawImage(crossword.getImage(), 0, 0, IMAGE_WIDTH_IN_POINTS, targetImageHeight, null);

            // Compute where the crossword grid will begin horizontally.
            int originX = IMAGE_WIDTH_IN_POINTS + 36;
            printGrid(graphics, pageFormat, originX, 0, false);

            // Add the rectangles occupied by the grid and the photograph
            // to restrictedList.
            restrictedList.add(new Rectangle(originX, 0, //
                    SQUARE_SIDE * crossword.getGrid().getColCount(), SQUARE_SIDE * crossword.getGrid().getRowCount()));
            restrictedList.add(new Rectangle(0, 0, //
                    IMAGE_WIDTH_IN_POINTS, targetImageHeight));

            // Repeat the process of trying to fit all the clues on
            // the page, trying successively smaller fonts until
            // successful.
            Font clueFont = new Font(Font.SERIF, Font.PLAIN, 10);
            while (true) {
                if (addClues(clueFont, graphics, pageFormat, restrictedList)) {
                    break;
                }
                float clueFontSize = ((Double) (clueFont.getSize2D() - 0.5)).floatValue();
                clueFont = clueFont.deriveFont(clueFontSize);
            }
            return PAGE_EXISTS;
        }

        // For page 2, print the solution.
        case 1: {
            printGrid(graphics, pageFormat, 0, 0, true);
            return PAGE_EXISTS;
        }
        default: {
            return NO_SUCH_PAGE;
        }
        }
    }

    private void printGrid(Graphics g, PageFormat pageFormat, int originX, int originY, boolean printSolution) {
        Font font = null;
        if (printSolution) {
            font = new Font(Font.SANS_SERIF, Font.PLAIN, 10);
        } else {
            font = new Font(Font.SANS_SERIF, Font.PLAIN, 5);
        }
        g.setFont(font);
        FontMetrics metrics = g.getFontMetrics();

        Grid grid = crossword.getGrid();
        int y = originY;
        for (int row = 0; row < grid.getRowCount(); row++) {
            int x = originX;
            for (int column = 0; column < grid.getColCount(); column++) {
                Square square = grid.getSquare()[row][column];
                g.drawRect(x, y, SQUARE_SIDE, SQUARE_SIDE);
                if (square.getValue() == null) {
                    g.fillRect(x, y, SQUARE_SIDE, SQUARE_SIDE);
                } else {
                    if (printSolution) {
                        String value = square.getValue().toString();
                        int charWidth = metrics.stringWidth(value);
                        int xPosition = x + (SQUARE_SIDE - charWidth) / 2;
                        g.drawString(value, xPosition, y + metrics.getHeight() - 2);
                    } else {
                        if (square.getClueNumber() != null) {
                            g.drawString(square.getClueNumber().toString(), //
                                    x + 1, y + metrics.getHeight() - 2);
                        }
                    }
                }
                x = x + SQUARE_SIDE;
            }
            y = y + SQUARE_SIDE;
        }
    }

    // Add the clues to the grid, if they fit. If they do, return true.
    private boolean addClues(Font clueFont, Graphics graphics, PageFormat pageFormat, List<Rectangle> restrictedList) {
        graphics.setFont(clueFont);

        // Start by dividing the page into five columns.
        int columnCount = 5;
        int columnWidth = ((Double) (pageFormat.getImageableWidth() / columnCount)).intValue();

        // For each column, determine how much space there is between
        // any restricted area, or the top of the page, and the bottom
        // of the page. Then add the number of lines that will fit to
        // a running total. We won't try to put clues between the two
        // graphics.
        int linesAvailable = 0;
        int lineHeight = graphics.getFontMetrics().getHeight();
        for (int column = 0; column < columnCount; column++) {
            int x = column * columnWidth;
            int columnTop = clueColumnStartY(x, columnWidth, lineHeight, pageFormat, restrictedList);
            double spaceAvailable = pageFormat.getImageableHeight() - columnTop;
            Double linesInColumn = spaceAvailable / lineHeight;
            linesAvailable += linesInColumn.intValue();
        }

        // Now see if the clues will fit.
        List<String> clueTextList = new ArrayList<String>();
        for (String clueText : clueList) {
            // Split the clue text into as many lines as necessary
            // and add them to clueTextList. We'll leave a cushion
            // of 7 points between columns.
            // If divideClue couldn't succeed, return false.
            List<String> dividedLines = divideClue(clueText, clueFont, graphics, columnWidth - 7);
            if (dividedLines.isEmpty()) {
                return false;
            }
            clueTextList.addAll(dividedLines);
        }

        // If the number if lines required exceeds the number available,
        // return false. The caller will try again with a smaller font.
        if (clueTextList.size() > linesAvailable) {
            return false;
        }

        // The lines will fit in the available space. Add them
        // to the page.
        int x = 0;

        // Determine the starting y for the first column
        // from x and the values in restrictedList.
        // Increment y after each line. When the next
        // y would exceed the page bottom, add columnWidth
        // to x and compute y for the top of the next column.
        int y = clueColumnStartY(x, columnWidth, lineHeight, pageFormat, restrictedList);
        while (!clueTextList.isEmpty()) {
            String clueText = clueTextList.remove(0);
            graphics.drawString(clueText, x, y);
            y += lineHeight;
            if (y > pageFormat.getImageableHeight()) {
                x += columnWidth;
                y = clueColumnStartY(x, columnWidth, lineHeight, pageFormat, restrictedList);
            }
        }

        return true;
    }

    // Get the y position for the first clue in a column.
    private int clueColumnStartY(int x, int columnWidth, int lineHeight, PageFormat pageFormat,
            List<Rectangle> restrictedList) {
        int columnTop = lineHeight;
        for (Rectangle rectangle : restrictedList) {
            // If x or (x + columnWidth) is between the minimum
            // and maximum x for the rectangle, ensure that
            // columnTop is below the bottom of the rectangle.
            boolean xContained = false;
            if (x >= rectangle.getMinX() && x <= rectangle.getMaxX()) {
                xContained = true;
            }
            if ((x + columnWidth) >= rectangle.getMinX() && (x + columnWidth) <= rectangle.getMaxX()) {
                xContained = true;
            }
            if (xContained) {
                if ((rectangle.getY() + rectangle.getHeight()) > columnTop) {
                    // Calculate columnTop as the bottom of the rectangle
                    // plus 0.25 in.
                    columnTop = ((Double) (rectangle.getY() + rectangle.getHeight() + 18)).intValue();
                }
            }
        }

        return columnTop;
    }

    // Split the clue text into as many lines as necessary to fit within
    // columnWidth.
    private List<String> divideClue(String clueText, Font clueFont, Graphics graphics, int columnWidth) {
        List<String> lines = new ArrayList<String>();
        StringBuilder textBuf = new StringBuilder(clueText);

        // While the width of the clue exceeds the available width,
        // break off words until the next word would exceed the width.
        while (graphics.getFontMetrics().stringWidth(textBuf.toString()) > columnWidth) {
            // If there's a space left over at the start of the buffer,
            // remove it.
            if (textBuf.charAt(0) == ' ') {
                textBuf.deleteCharAt(0);
            }

            // lineText contains the words as they're added
            // to the line.
            StringBuilder lineText = new StringBuilder();

            // Start each line after the first with a space.
            if (!lines.isEmpty()) {
                lineText.append(" ");
            }

            boolean full = false;
            String space = "";

            while (!full) {
                // Find the next space following the first character.
                int spacePos = textBuf.indexOf(" ", 1);

                // If there was no space, set spacePos to the
                // length of textBuf.
                if (spacePos < 0) {
                    spacePos = textBuf.length();
                }

                // The next word is everything up to but not
                // including the space.
                String nextWord = textBuf.substring(0, spacePos);

                // If the current lineText plus a space plus
                // nextWord exceeds the column width, we're
                // done.
                // Otherwise, append nextWord to lineText and
                // delete it from textBuf.
                if (graphics.getFontMetrics().stringWidth(lineText.toString() + " " + nextWord) > columnWidth) {
                    full = true;
                } else {
                    lineText.append(space + nextWord);
                    textBuf.delete(0, spacePos + 1);
                    space = " ";
                }
            }

            // If the text to be added is empty, return an empty list.
            if (lineText.toString().isBlank()) {
                lines.clear();
                return lines;
            }
            // Add the accumulated lineText to the return list.
            lines.add(lineText.toString());
        }

        // When what's left of the clue fits within columnWidth,
        // add it to lines.
        if (!lines.isEmpty()) {
            textBuf.insert(0, ' ');
        }
        lines.add(textBuf.toString());

        return lines;
    }
}
